/* 
 * vmsdevchan.c
 *
 *	Channel driver for VMS channels based on devices
 *
 */

#include "tclInt.h"	/* Internal definitions for Tcl. */
#include "tclPort.h"	/* Portability features for Tcl. */
#include <unixio.h>
#include <unixlib.h>

#include <errno.h>
#include <ssdef.h>
#include <iodef.h>
#include <starlet.h>
#include <devdef.h>
#include <dvidef.h>
#include <msgdef.h>

typedef struct {
    int	l;
    void *a;
} DESC;

#define	LOAD_DESC(d, s) d.l=strlen(s);d.a=s

/*
 * This structure describes per-instance state of a device based channel.
 */

typedef struct DevState {
    int flags;
    Tcl_File dev;
    int chan;
    int eventFlag;		/* Set when mbox I/O completes */
    int checkMask;		/* what we're checking */
    int readyMask;		/* what's ready */
    short iosb[4];
    int mbxChan;		/* VMS mailbox channel for unsolicited input */
} DevState;

/*
 * Bit values for flags member of DevState
 */

#define	DEV_IS_TERM	(1<<0)
#define	DEV_IS_MBX	(1<<1)
#define DEV_WATCHING	(1<<2)

/*
 * Static routines for this file:
 */

static DevState *	CreateDev _ANSI_ARGS_((Tcl_Interp *interp,
			    char *name, char *mode, int perm));
static int		DevBlockModeProc _ANSI_ARGS_((
			    ClientData instanceData, int mode));
static int		DevCloseProc _ANSI_ARGS_((ClientData instanceData,
			    Tcl_Interp *interp));
static int		DevInputProc _ANSI_ARGS_((ClientData instanceData,
		            char *buf, int toRead, int *errorCode));
static int		DevOutputProc _ANSI_ARGS_((
			    ClientData instanceData, char *buf, int toWrite,
			    int *errorCode));
static void		DevWatchProc _ANSI_ARGS_((ClientData instanceData,
			    int mask));
static int		DevReadyProc _ANSI_ARGS_((ClientData instanceData,
			    int mask));
static Tcl_File		DevGetProc _ANSI_ARGS_((ClientData instanceData,
			    int direction));
static void		devFreeProc(ClientData);
static void		setVMSerrorCode(int);
static int		assignChan(char *, char *);
static int		getDeviceType(int);
static int		getDeviceName(int, char *);
static int		crembx(char *);
static int		delmbx(int);
static int		cleanmbx(int);
static void		devReadyAstProc(ClientData);
static int		setWriteAttention(int, void (*)(ClientData), ClientData);

static void debugOut(char *);

/*
 * This structure describes the channel type structure for device based IO:
 */

static Tcl_ChannelType devChannelType = {
    "dev",				/* Type name. */
    DevBlockModeProc,			/* Set blocking/nonblocking mode.*/
    DevCloseProc,			/* Close proc. */
    DevInputProc,			/* Input proc. */
    DevOutputProc,			/* Output proc. */
    NULL,				/* Seek proc. */
    NULL,				/* Set option proc. */
    NULL,				/* Get option proc. */
    DevWatchProc,			/* Initialize notifier. */
    DevReadyProc,			/* Are there events? */
    DevGetProc,				/* Get Tcl_Files out of channel. */
};


/*
 *----------------------------------------------------------------------
 *
 * DevBlockModeProc --
 *
 *	Set blocking or non-blocking mode on channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
DevBlockModeProc(instanceData, mode)
    ClientData instanceData;		/* Unused. */
    int mode;				/* The mode to set. */
{
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * DevInputProc --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *	This routine is used by the file and command pipeline drivers.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
DevInputProc(instanceData, buf, toRead, errorCode)
    ClientData instanceData;		/* Unused. */
    char *buf;				/* Where to store data read. */
    int toRead;				/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    short iosb[4];
    int chan, status;			/* The OS handle for reading. */
    int bytesRead;			/* How many bytes were actually
                                         * read from the input device? */
    DevState *statePtr = (DevState *)instanceData;

    *errorCode = 0;
    errno = 0;
    chan = (int) Tcl_GetFileInfo(statePtr->dev, NULL);
    
    /*
     * Assume there is always enough input available. This will block
     * appropriately, and read will unblock as soon as a short read is
     * possible, if the channel is in blocking mode. If the channel is
     * nonblocking, the read will never block.
     */
    
    if (toRead == 0) {
        return 0;
    }

    statePtr->readyMask &= ~TCL_READABLE;

    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
		      buf, toRead, 0, 0, 0, 0);
    if ((status & 1) && ((status = iosb[0]) & 1)) {
	if (statePtr->flags & DEV_IS_TERM) {
	    if (iosb[1] == 0 && iosb[3] == 1 && iosb[2] == 26)
		return 0;
	    else
		iosb[1] += iosb[3];	/* Add in terminator length */
	}
	return iosb[1];
    }
    else {
	setVMSerrorCode(status);
	*errorCode = errno;
	return -1;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DevOutputProc--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static char	*newBuff = 0;
static int	newBuffSize = 4096*2;

static int
DevOutputProc(instanceData, buf, toWrite, errorCode)
    ClientData instanceData;		/* Unused. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    char *b;
    short iosb[4];
    int written = 0;
    int chan, status;
    int nWrite;
    DevState *statePtr = (DevState *) instanceData;

    *errorCode = 0;
    errno = 0;
    chan = (int) Tcl_GetFileInfo(statePtr->dev, NULL);

    /*
     * Terminals are special: need to add CR's to each LF output
     */
    if (statePtr->flags & DEV_IS_TERM) {
	char *in, *out, c;
	int i;

	if (newBuff == (char *)NULL)
	    newBuff = ckalloc(newBuffSize);

	if (toWrite > newBuffSize/2) {
	    newBuff = realloc(newBuff, toWrite*2);
	    newBuffSize = toWrite*2;
	}

	in = buf;
	b = out = newBuff;
	for (i=0, nWrite=0; i<toWrite; i++) {
	    c = *out++ = *in++;
	    nWrite++;
	    written++;

#if 0
	    if (nWrite >= sizeof(newBuff)) {
		status = sys$qiow(0, chan, IO$_WRITEVBLK, iosb, 0, 0,
				  b, nWrite, 0, 0, 0, 0);
		if ((status & 1) && ((status = iosb[0]) & 1)) {
		    nWrite = 0;
		    out = newBuff;
		}
		else {
		    setVMSerrorCode(status);
		    *errorCode = errno;
		    return -1;
		}
	    }
#endif
	    if (c == '\r') {
		*out++ = '\n';
		nWrite++;
	    }
	}
    }
    else {
	b = buf;
	nWrite = toWrite;
    }
    status = sys$qiow(0, chan, IO$_WRITEVBLK, iosb, 0, 0,
			  b, nWrite, 0, 0, 0, 0);
    if ((status & 1) && ((status = iosb[0]) & 1)) {
	if ((statePtr->flags & DEV_IS_TERM) && nWrite == iosb[1])
	    return written;
	else
	    return iosb[1];
    }
    else {
	setVMSerrorCode(status);
	*errorCode = errno;
	return -1;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DevCloseProc --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

static int
DevCloseProc(instanceData, interp)
    ClientData instanceData;	/* Unused. */
    Tcl_Interp *interp;
{
    int chan, errorCode = 0;
    int status;
    DevState *statePtr = (DevState *) instanceData;

    if (statePtr->dev != NULL) {
        chan = (int) Tcl_GetFileInfo(statePtr->dev, NULL);

#if 0
	/*
	 * Check for read/write file so we only close it once.
	 */

	if (inFile == outFile) {
	    outFile = NULL;
	}
#endif
        Tcl_FreeFile(statePtr->dev);
	if (sys$dassgn(chan) < 0) {
	    errorCode = errno;
	}
    }

#if 0
    if (outFile != NULL) {
        chan = (int) Tcl_GetFileInfo(outFile, NULL);
        Tcl_FreeFile(outFile);        
	status = sys$dassgn(chan);
	if (!(status & 1) && errorCode == 0) {
	    errorCode = errno;
        } 
    }
#endif
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * DevWatchProc --
 *
 *	Initialize the notifier to watch Tcl_Files from this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Sets up the notifier so that a future event on the channel will
 *	be seen by Tcl.
 *
 *----------------------------------------------------------------------
 */

static void
DevWatchProc(instanceData, mask)
    ClientData instanceData;		/* The file state. */
    int mask;				/* Events of interest; an OR-ed
    					 * combination of TCL_READABLE,
    					 * TCL_WRITABLE and TCL_EXCEPTION. */
{
    DevState *statePtr = (DevState *) instanceData;

    Tcl_WatchFile(statePtr->dev, mask);
}

/*
 *----------------------------------------------------------------------
 *
 * DevReadyProc --
 *
 *	Called by the notifier to check whether events of interest are
 *	present on the channel.
 *
 * Results:
 *	Returns OR-ed combination of TCL_READABLE, TCL_WRITABLE and
 *	TCL_EXCEPTION to indicate which events of interest are present.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
DevReadyProc(instanceData, mask)
    ClientData instanceData;		/* The file state. */
    int mask;				/* Events of interest; an OR-ed
                                         * combination of TCL_READABLE,
                                         * TCL_WRITABLE and TCL_EXCEPTION. */
{
    DevState *statePtr = (DevState *) instanceData;

    return Tcl_FileReady(statePtr->dev, mask);
}

/*
 *----------------------------------------------------------------------
 *
 * DevGetProc --
 *
 *	Called from Tcl_GetChannelFile to retrieve Tcl_Files from inside
 *	a VMS device based channel.
 *
 * Results:
 *	The appropriate Tcl_File or NULL if not present. 
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static Tcl_File
DevGetProc(instanceData, direction)
    ClientData instanceData;		/* The socket state. */
    int direction;			/* Which Tcl_File to retrieve? */
{
    DevState *statePtr = (DevState *) instanceData;

    return statePtr->dev;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenDevChannel --
 *
 *	Open an device based channel on VMS systems.
 *
 * Results:
 *	The new channel or NULL. If NULL, the output argument
 *	errorCodePtr is set to a POSIX error.
 *
 * Side effects:
 *	May open the channel
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenDevChannel(Tcl_Interp *interp, char *fileName, char *modeString, int permissions, char *chanName)
{
    char *s;
    char cName[20];
    char *channelName;
    int status;
    Tcl_Channel chan;
    DevState *statePtr;
    int mode;

    statePtr = CreateDev(interp, fileName, modeString, permissions);
    if (statePtr == NULL)
	return NULL;

    if (chanName)
	channelName = chanName;
    else {
	sprintf(cName, "dev%d", (int)Tcl_GetFileInfo(statePtr->dev, NULL));
	channelName = cName;
    }
    
    mode = 0;
    if (strchr(modeString, 'r'))
    	mode |= TCL_READABLE;
    if (strchr(modeString, 'w'))
    	mode |= TCL_WRITABLE;
    chan = Tcl_CreateChannel(&devChannelType, channelName,
            (ClientData) statePtr, mode);

    if (statePtr->flags & DEV_IS_TERM) {
	Tcl_SetChannelOption(interp, chan, "-translation", "cr");
	Tcl_SetChannelOption(interp, chan, "-buffering", "line");
    }
    return chan;
}

static DevState *
CreateDev(Tcl_Interp *interp, char *name, char *modeString, int perm)
{
    char mbxName[65], *mbn;
    int status, mbxChan;
    int channel, seekFlag, mode, channelPermissions;
    DevState *statePtr;

    mode = TclGetOpenMode(interp, modeString, &seekFlag);
    if (mode == -1) {
        return NULL;
    }
    switch (mode & (O_RDONLY | O_WRONLY | O_RDWR)) {
	case O_RDONLY:
	    channelPermissions = TCL_READABLE;
	    break;
	case O_WRONLY:
	    channelPermissions = TCL_WRITABLE;
	    break;
	case O_RDWR:
	    channelPermissions = (TCL_READABLE | TCL_WRITABLE);
	    break;
	default:
	    Tcl_AppendResult(interp, "invalid mode value", (char *)NULL);
	    return NULL;
    }

    /*
     * 1. Create mailbox
     * 2. Get name of mailbox
     * 3. Create channel w/ mbox
     * 4. Set write attention on mbox
     */
    if (channelPermissions & TCL_READABLE) {
	mbxChan = crembx(mbxName);
	if (!mbxChan) {
	    if (interp) {
		Tcl_AppendResult(interp, "couldn't create a mailbox: ",
			Tcl_PosixError(interp), (char *) NULL);
	    }
	    return NULL;
	}
	mbn = mbxName;
    }
    else {
    	mbn = 0;
    	mbxChan = 0;
    }
    channel = assignChan(name, mbn);
    if (channel == 0) {
    	if (interp) {
	    Tcl_AppendResult(interp, "couldn't open \"", name, "\": ",
		    Tcl_PosixError(interp), (char *) NULL);
	}
	if (mbxChan)
	    delmbx(mbxChan);
	return NULL;
    }
    statePtr = (DevState *) ckalloc((unsigned) sizeof(DevState));
    statePtr->dev = Tcl_GetFile((ClientData) channel, TCL_VMS_DEVICE);
    statePtr->chan = channel;
    statePtr->flags = 0;
    statePtr->eventFlag = getEventFlag();
    statePtr->checkMask = 0;
    statePtr->readyMask = 0;
    statePtr->mbxChan = mbxChan;

    /*
     * Figure out what kind of device this is
     */
    status = getDeviceType(channel);
    if (status < 0) {
    	if (interp) {
	    Tcl_AppendResult(interp, "couldn't get device information:",
		    Tcl_PosixError(interp), (char *) NULL);
	}
	devFreeProc((ClientData)statePtr);
	return NULL;
    }
    else if (status == DEV_IS_TERM)
	statePtr->flags |= DEV_IS_TERM;
    else if (status == DEV_IS_MBX)
	statePtr->flags |= DEV_IS_TERM;
    else {
	/* who knows */
    }

    if (statePtr->flags & DEV_IS_TERM || statePtr->flags & DEV_IS_MBX) {
	/*
	 * Issue a "set write attention AST" to let us know
	 * when data is available to be read.
	 */
	if (channelPermissions & TCL_READABLE) {
	    if (!setWriteAttention(mbxChan, devReadyAstProc, statePtr)) {
	    	if (interp) {
		    Tcl_AppendResult(interp,
			"couldn't set write attention on device \"", name,
			"\":", Tcl_PosixError(interp), (char *) NULL);
		}
		devFreeProc((ClientData)statePtr);
		return 0;
	    }
	}
    }

    Tcl_SetNotifierData(statePtr->dev, devFreeProc, (ClientData)statePtr);
    return statePtr;
}

static void
devReadyAstProc(ClientData clientdata)
{
    /*
     * AST handler invoked when there's data to be read
     */
    DevState *statePtr = (DevState *)clientdata;

    /*
     * Read the notification message that triggered the AST
     */
    cleanmbx(statePtr->mbxChan);

    if (statePtr->checkMask & TCL_READABLE) {
	statePtr->readyMask |= TCL_READABLE;
	if (statePtr->flags & DEV_WATCHING) {
	    sys$setef(statePtr->eventFlag);	/* For vmsselect */
	}
    }
    setWriteAttention(statePtr->mbxChan, devReadyAstProc, statePtr);
}

static void
devFreeProc(ClientData clientdata)
{
    /*
     * Free up device info structure
     */
    DevState *statePtr = (DevState *)clientdata;

    if (statePtr) {
	if (statePtr->eventFlag)
	    freeEventFlag(statePtr->eventFlag);
	if (statePtr->mbxChan)
	    delmbx(statePtr->mbxChan);
	ckfree((char *)statePtr);
    }
}

TclVmsWatchDevice(Tcl_File file, int mask)
{
    /*
     * Watch for something to happen on a device.
     */
    DevState *statePtr = (DevState *)Tcl_GetNotifierData(file, NULL);

    sys$setast(0);
    if (statePtr->flags & (DEV_IS_TERM | DEV_IS_MBX)) {
	statePtr->checkMask = mask;
	if (mask & TCL_READABLE) {
	    if (statePtr->readyMask & TCL_READABLE) {
		sys$setef(statePtr->eventFlag);
	    }
	    else {
		sys$clref(statePtr->eventFlag);
		statePtr->flags |= DEV_WATCHING;
	    }
	    sys$setast(1);
	    return statePtr->eventFlag;
	}
    }
    sys$setast(1);
    return 0;
}

VmsDeviceReady(Tcl_File file, int mask)
{
    /*
     * Check and see what's ready for a device
     */
    int readyMask = 0;
    DevState *statePtr = (DevState *)Tcl_GetNotifierData(file, NULL);

    sys$setast(0);
    if (statePtr->readyMask & TCL_READABLE)
	readyMask |= TCL_READABLE;
    if (statePtr->readyMask & TCL_WRITABLE)
	readyMask |= TCL_WRITABLE;
    if (statePtr->readyMask & TCL_EXCEPTION)
	readyMask |= TCL_EXCEPTION;
    statePtr->flags |= DEV_WATCHING;
    /* readyMask = 0; */
    sys$setast(1);
    return readyMask;
}

int
Tcl_DevCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    /*
     * TCL command for doing I/O on VMS devices.
     *
     * [dev deviceName ...]
     */
    char *modeString;
    int prot;
    Tcl_Channel chan;

    if (argc != 2 || argc > 4) {
	Tcl_AppendResult(interp, "Wrong # of arguments to \"dev\" command",
			 (char *)NULL);
	return TCL_ERROR;
    }
    prot = 0666;
    if (argc == 2) {
	modeString = "r";
    }
    else  {
	modeString = argv[2];
	if (argc == 4) {
	    if (Tcl_GetInt(interp, argv[3], &prot) != TCL_OK) {
		return TCL_ERROR;
	    }
	}
    }
    chan = Tcl_OpenDevChannel(interp, argv[1], modeString, prot, (char *)0);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    Tcl_RegisterChannel(interp, chan);
    Tcl_AppendResult(interp, Tcl_GetChannelName(chan), (char *) NULL);
    return TCL_OK;
}

static void
setVMSerrorCode(int status)
{
    /*
     * Convert a VMS error number to a C error number
     */
#ifdef EVMSERR
    errno = EVMSERR;
#else
    errno = VMSERRNO;
#endif	/* EVMSERR */
    vaxc$errno = status;
}

static
assignChan(char *name, char *mbxName)
{
    short channel;
    int status;
    DESC d, mbxD;

    LOAD_DESC(d, name);
    if (mbxName) {
	LOAD_DESC(mbxD, mbxName);
	status = sys$assign(&d, &channel, 0, &mbxD, 0);
    }
    else
	status = sys$assign(&d, &channel, 0, 0, 0);
    if (status & 1)
	return channel;
    else {
    	printf("assignChan for %s failed: %x\n", name, status);
	setVMSerrorCode(status);
        return 0;
    }
}

static
crembx(char *name)
{
    /*
     * Create a mailbox and return its name.
     */
    short chan;
    int status;

    status = sys$crembx(0, &chan, 0, 0, 0, 0, 0, 0);
    if (status & 1) {
	if (getDeviceName(chan, name))
	    return chan;
	else {
	    delmbx(chan);
	    return 0;
	}
    }
    else {
	setVMSerrorCode(status);
	return 0;
    }
}

static
delmbx(int chan)
{
    int status;

    status = sys$delmbx(chan);
    if (status & 1)
	return 1;
    else {
	setVMSerrorCode(status);
	return 0;
    }
}

static
cleanmbx(int chan)
{
    /*
     * Read a message in our mailbox
     */
    char buff[512];
    short iosb[4];
    short brdcstCount;
    int status;
    struct {
	short msgType;
	short unitNumber;
	char count;
	char controller[15];
    } msg;

    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
		      &msg, sizeof(msg), 0, 0, 0, 0);
    if ((status & 1) && ((status = iosb[0]) & 1)) {
	if (msg.msgType == MSG$_TRMBRDCST) {
	    /*
	     * Suck up broadcast message
	     */
	    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
			      &brdcstCount, 2, 0, 0, 0, 0);
	    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
			      buff, &brdcstCount, 0, 0, 0, 0);
	}
    }
    else
	printf("Could not clean mailbox: %x\n", status);
}

static
getDeviceType(int chan)
{
    int status;
    int retLength;
    struct {
	short l;
	short code;
	void *a;
	int *retLength;
    } dviItemList[2];
    union devdef deviceChar;

    dviItemList[0].l = sizeof(deviceChar);
    dviItemList[0].code = DVI$_DEVCHAR;
    dviItemList[0].a = &deviceChar;
    dviItemList[0].retLength = &retLength;
    dviItemList[1].l = dviItemList[1].code = 0;
    status = sys$getdviw(0, chan, 0, &dviItemList, 0, 0, 0, 0);
    if (status & 1) {
	if (deviceChar.dev$v_trm)
	    return DEV_IS_TERM;
	else if (deviceChar.dev$v_mbx)
	    return DEV_IS_MBX;
	else
	    return 0;
    }
    else {
	setVMSerrorCode(status);
	return -1;
    }
}

static
getDeviceName(int chan, char *name)
{
    int status;
    int retLength = 0;
    struct {
	short l;
	short code;
	void *a;
	int *retLength;
    } dviItemList[2];

    dviItemList[0].l = 64;
    dviItemList[0].code = DVI$_DEVNAM;
    dviItemList[0].a = name;
    dviItemList[0].retLength = &retLength;
    dviItemList[1].l = dviItemList[1].code = 0;
    status = sys$getdviw(0, chan, 0, &dviItemList, 0, 0, 0, 0);
    if (status & 1) {
	name[retLength] = '\0';
	return  1;
    }
    else {
	setVMSerrorCode(status);
	return 0;
    }
}

static int
setWriteAttention(int chan, void (*devReadyAstProc)(ClientData), ClientData clientdata)
{
    short iosb[4];
    int status;

    status = sys$qiow(0, chan, IO$_SETMODE|IO$M_WRTATTN, iosb, 0, 0,
		      devReadyAstProc, clientdata, 0, 0, 0, 0);
    return 1;
}

static short outfd = -1;

static void
debugOut(char *msg)
{
    if (outfd == -1) {
	outfd = assignChan("sys$output", (char *)0);
    }
    sys$qiow(0, outfd, IO$_WRITEVBLK, 0, 0, 0,
			  "\n", 1, 0, 0, 0, 0);
    sys$qiow(0, outfd, IO$_WRITEVBLK, 0, 0, 0,
			  msg, strlen(msg), 0, 0, 0, 0);
    sys$qiow(0, outfd, IO$_WRITEVBLK, 0, 0, 0,
			  "\r\n", 2, 0, 0, 0, 0);
}
